`timescale 1ns / 1ps
`default_nettype none

//******************************************************************************
//* Az UART funkcit megvalst modul.                                        *
//******************************************************************************
module uart(
   //rajel s reset.
   input  wire                 clk,            //Rendszerrajel
   input  wire                 rst,            //Aktv magas szinkron reset
   
   //Regiszter rsi interfsz.
   input  wire [4:0]           wr_addr,        //rsi cm
   input  wire                 wr_en,          //rs engedlyez jel
   input  wire [31:0]          wr_data,        //rsi adat
   input  wire [3:0]           wr_strb,        //Bjt engedlyez jelek
    
   //Regiszter olvassi interfsz.
   input  wire [4:0]           rd_addr,        //Olvassi cm
   input  wire                 rd_en,          //Olvass engedlyez jel
   output reg  [31:0]          rd_data,        //Olvassi adat
   
   //UART adatvonalak.
   input  wire                 uart_rxd,       //Soros bemenet
   output wire                 uart_txd,       //Soros kimenet
   
   //Megszaktskr kimenet.
   output reg                  irq
);

//******************************************************************************
//* Vezrl regiszter.                  BZIS+0x00, 32 bites, rhat/olvashat *
//******************************************************************************
wire ctrl_reg_wr = wr_en & (wr_addr[4:2] == 3'd0) & (wr_strb == 4'b1111);
wire ctrl_reg_rd = rd_en & (rd_addr[4:2] == 3'd0);

//Ads engedlyezs (0. bit).
reg txen;

always @(posedge clk)
begin
   if (rst)
      txen <= 1'b0;
   else
      if (ctrl_reg_wr)
         txen <= wr_data[0];
end

//Vtel engedlyezs (1. bit).
reg rxen;

always @(posedge clk)
begin
   if (rst)
      rxen <= 1'b0;
   else
      if (ctrl_reg_wr)
         rxen <= wr_data[1];
end

//Adsi FIFO trlse (2. bit).
wire txclr = ctrl_reg_wr & wr_data[2];

//Vteli FIFO trlse (3. bit).
wire rxclr = ctrl_reg_wr & wr_data[3];

//A regiszterbl visszaolvashat rtk.
wire [31:0] ctrl_reg_dout = {30'd0, rxen, txen};


//******************************************************************************
//* rajel oszts regiszter.            BZIS+0x04, 32 bites, rhat/olvashat *
//* baud rate = f_sysclk / (16 * (CLKDIV + 1))                                 *
//******************************************************************************
wire clkdiv_reg_wr = wr_en & (wr_addr[4:2] == 3'd1) & (wr_strb == 4'b1111);
wire clkdiv_reg_rd = rd_en & (rd_addr[4:2] == 3'd1);

//rajel oszts (0 - 15. bitek).
reg [15:0] clkdiv;

always @(posedge clk)
begin
   if (rst)
      clkdiv <= 16'd0;
   else
      if (clkdiv_reg_wr)
         clkdiv <= wr_data[15:0];
end

//A regiszterbl visszaolvashat rtk.
wire [31:0] clkdiv_reg_dout = {16'd0, clkdiv};


//******************************************************************************
//* Vonali kdols regiszter.           BZIS+0x08, 32 bites, rhat/olvashat *
//******************************************************************************
wire coding_reg_wr = wr_en & (wr_addr[4:2] == 3'd2) & (wr_strb == 4'b1111);
wire coding_reg_rd = rd_en & (rd_addr[4:2] == 3'd2);

//Adatbitek szma (0 - 1. bitek).
reg [1:0] databits;

always @(posedge clk)
begin
   if (rst)
      databits <= 2'b11;
   else
      if (coding_reg_wr)
         databits <= wr_data[1:0];
end

//Parits engedlyezs (2. bit)
reg paren;

always @(posedge clk)
begin
   if (rst)
      paren <= 1'b0;
   else
      if (coding_reg_wr)
         paren <= wr_data[2];
end

//Parits tpus (3 - 4. bitek).
reg [1:0] partype;

always @(posedge clk)
begin
   if (rst)
      partype <= 2'b00;
   else
      if (coding_reg_wr)
         partype <= wr_data[4:3];
end

//STOP bitek szma (5. bit).
reg stopbits;

always @(posedge clk)
begin
   if (rst)
      stopbits <= 1'b0;
   else
      if (coding_reg_wr)
         stopbits <= wr_data[5];
end

//A regiszterbl visszaolvashat rtk.
wire [31:0] coding_reg_dout = {26'd0, stopbits, partype, paren, databits};


//******************************************************************************
//* FIFO sttusz regiszter.               BZIS+0x0C, 32 bites, csak olvashat *
//******************************************************************************
wire status_reg_rd = rd_en & (rd_addr[4:2] == 3'd3);

wire tx_exists;
wire tx_full;
wire rx_exists;
wire rx_full;

//A regiszterbl visszaolvashat rtk.
wire [31:0] status_reg_dout = {28'd0, rx_full, rx_exists, ~tx_exists, ~tx_full};


//******************************************************************************
//* Adatregiszter.                      BZIS+0x10, 32 bites, rhat/olvashat *
//******************************************************************************
wire data_reg_wr = wr_en & (wr_addr[4:2] == 3'd4) & (wr_strb == 4'b1111);
wire data_reg_rd = rd_en & (rd_addr[4:2] == 3'd4);

wire [8:0] rx_fifo_dout;

//A regiszterbl visszaolvashat rtk.
wire [31:0] data_reg_dout = {23'd0, rx_fifo_dout};


//******************************************************************************
//* Megszakts eng. reg.               BZIS+0x14, 32 bites, rhat/olvashat *
//******************************************************************************
wire ie_reg_wr = wr_en & (wr_addr[4:2] == 3'd5) & (wr_strb == 4'b1111);
wire ie_reg_rd = rd_en & (rd_addr[4:2] == 3'd5);

//FIFO sttusz megszaktsok engedlyezse (0 - 3. bitek).
reg [3:0] ie_reg;

always @(posedge clk)
begin
   if (rst)
      ie_reg <= 4'b0000;
   else
      if (ie_reg_wr)
         ie_reg <= wr_data[3:0];
end

//A regiszterbl visszaolvashat rtk.
wire [31:0] ie_reg_dout = {28'd0, ie_reg};

//Az aktv magas szintrzkeny megszaktskr kimenet meghajtsa.
always @(posedge clk)
begin
   irq <= |(status_reg_dout[3:0] & ie_reg);
end


//******************************************************************************
//* Olvassi multiplexer.                                                      *
//******************************************************************************
wire [5:0] rd_sel;

assign rd_sel[0] = ctrl_reg_rd;
assign rd_sel[1] = clkdiv_reg_rd;
assign rd_sel[2] = coding_reg_rd;
assign rd_sel[3] = status_reg_rd;
assign rd_sel[4] = data_reg_rd;
assign rd_sel[5] = ie_reg_rd;

always @(*)
begin
   case (rd_sel)
      6'b000001: rd_data <= ctrl_reg_dout;
      6'b000010: rd_data <= clkdiv_reg_dout;
      6'b000100: rd_data <= coding_reg_dout;
      6'b001000: rd_data <= status_reg_dout;
      6'b010000: rd_data <= data_reg_dout;
      6'b100000: rd_data <= ie_reg_dout;
      default  : rd_data <= 32'd0;
   endcase
end


//******************************************************************************
//* UART ad.                                                                  *
//******************************************************************************
uart_transmitter uart_transmitter_i(
   //rajel s reset.
   .clk(clk),                          //Rendszerrajel
   .rst(rst),                          //Aktv magas szinkron reset
   
   //Ads engedlyets.
   .enable(txen),
   
   //rajel oszts.
   .clk_div(clkdiv),
   
   //Vonali kdols.
   .data_bits(databits),               //Adatbitek szma
   .par_enable(paren),                 //Parits engedlyezs
   .par_type(partype),                 //Parits tpusa
   .stop_bits(stopbits),               //Stop bitek szma
   
   //Adsi FIFO.
   .data_in(wr_data[7:0]),             //Adat bemenet
   .fifo_wr(data_reg_wr),              //rs engedlyez jel
   .fifo_clr(txclr),                   //FIFO trl jel
   .fifo_exists(tx_exists),            //A FIFO-ban van adat
   .fifo_full(tx_full),                //A FIFO tele van
   
   //Soros adat kimenet
   .uart_txd(uart_txd)
);


//******************************************************************************
//* UART vev.                                                                 *
//******************************************************************************
uart_receiver uart_receiver_i(
   //rajel s reset.
   .clk(clk),                          //Rendszerrajel
   .rst(rst),                          //Aktv magas szinkron reset
   
   //Vtel engedlyets.
   .enable(rxen),
   
   //rajel oszts.
   .clk_div(clkdiv),
   
   //Vonali kdols.
   .data_bits(databits),               //Adatbitek szma
   .par_enable(paren),                 //Parits engedlyezs
   .par_type(partype),                 //Parits tpusa
   .stop_bits(stopbits),               //Stop bitek szma
   
   //Vteli FIFO.
   .data_out(rx_fifo_dout),            //Adat kimenet
   .fifo_rd(data_reg_rd),              //Olvass engedlyez jel
   .fifo_clr(rxclr),                   //FIFO trl jel
   .fifo_exists(rx_exists),            //A FIFO-ban van adat
   .fifo_full(rx_full),                //A FIFO tele van
   
   //Soros adat bemenet
   .uart_rxd(uart_rxd)
);


endmodule

`default_nettype wire
