`timescale 1ns / 1ps
`default_nettype none

//******************************************************************************
//* UART vev modul.                                                           *
//******************************************************************************
module uart_receiver(
   //rajel s reset.
   input  wire        clk,             //Rendszerrajel
   input  wire        rst,             //Aktv magas szinkron reset
   
   //Vtel engedlyets.
   input  wire        enable,
   
   //rajel oszts.
   input  wire [15:0] clk_div,
   
   //Vonali kdols.
   input  wire [1:0]  data_bits,       //Adatbitek szma
   input  wire        par_enable,      //Parits engedlyezs
   input  wire [1:0]  par_type,        //Parits tpusa
   input  wire        stop_bits,       //Stop bitek szma
   
   //Vteli FIFO.
   output wire [8:0]  data_out,        //Adat kimenet
   input  wire        fifo_rd,         //Olvass engedlyez jel
   input  wire        fifo_clr,        //FIFO trl jel
   output wire        fifo_exists,     //A FIFO-ban van adat
   output wire        fifo_full,       //A FIFO tele van
   
   //Soros adat bemenet
   input  wire        uart_rxd
);

//******************************************************************************
//* A soros bemenet mintavtelezse s lefut l detektls.                   *
//******************************************************************************
reg  [2:0] rxd_samples;
wire       rxd_sync    = rxd_samples[1];
wire       rxd_falling = (rxd_samples[2:1] == 2'b10);

always @(posedge clk)
begin
   if (rst)
      rxd_samples <= 3'b111;
   else
      rxd_samples <= {rxd_samples[1:0], uart_rxd};
end


//******************************************************************************
//* Baud rate genertor.                                                       *
//******************************************************************************
reg  [15:0] clkdiv_cnt_l;
wire        clkdiv_cnt_l_tc = (clkdiv_cnt_l == 16'd0);
reg  [3:0]  clkdiv_cnt_h;
wire        clkdiv_cnt_h_tc = (clkdiv_cnt_h == 4'd15) & clkdiv_cnt_l_tc;
wire        clkdiv_en;

always @(posedge clk)
begin
   if ((clkdiv_en == 0) || clkdiv_cnt_l_tc)
      clkdiv_cnt_l <= clk_div;
   else
      clkdiv_cnt_l <= clkdiv_cnt_l - 16'd1;
end

always @(posedge clk)
begin
   if (clkdiv_en == 0)
      clkdiv_cnt_h <= 4'd0;
   else
      if (clkdiv_cnt_l_tc)
         clkdiv_cnt_h <= clkdiv_cnt_h + 4'd1;
end


//******************************************************************************
//* Zajszrs a szinkronizlt soros adat bemeneten. A bitid 16 rszre van     *
//* felosztva. A 6., a 8. s a 10. temben mintt vesznk s tbbsgi          *
//* szavazssal dntnk.                                                       *
//******************************************************************************
reg [2:0] filter_reg;
reg       rxd_filtered;

always @(posedge clk)
begin
   if (rst)
      filter_reg <= 3'b000;
   else
      if (clkdiv_cnt_l_tc)
      begin
         if (clkdiv_cnt_h == 4'd5) filter_reg[0] <= rxd_sync;
         if (clkdiv_cnt_h == 4'd7) filter_reg[1] <= rxd_sync;
         if (clkdiv_cnt_h == 4'd9) filter_reg[2] <= rxd_sync;
      end
end

always @(*)
begin
   case (filter_reg)
      3'b000 : rxd_filtered <= 1'b0;
      3'b001 : rxd_filtered <= 1'b0;
      3'b010 : rxd_filtered <= 1'b0;
      3'b011 : rxd_filtered <= 1'b1;
      3'b100 : rxd_filtered <= 1'b0;
      3'b101 : rxd_filtered <= 1'b1;
      3'b110 : rxd_filtered <= 1'b1;
      default: rxd_filtered <= 1'b1;
   endcase
end


//******************************************************************************
//* Vteli shiftregiszter.                                                     *
//******************************************************************************
reg  [7:0] rx_shr;
wire       rx_shr_en;

always @(posedge clk)
begin
   if (rst)
      rx_shr <= 8'd0;
   else
      if (rx_shr_en)
         rx_shr <= {rxd_filtered, rx_shr[7:1]};
end

//A shiftregiszter prhuzamos kimenete a vonali kdolsi belltsok alapjn.
reg [7:0] rx_data;

always @(*)
begin
   case (data_bits)
      2'b00  : rx_data <= {3'b000, rx_shr[7:3]};  //5 bit
      2'b10  : rx_data <= {2'b00,  rx_shr[7:2]};  //6 bit
      2'b01  : rx_data <= {1'b0,   rx_shr[7:1]};  //7 bit
      default: rx_data <=          rx_shr[7:0];   //8 bit
   endcase
end

//Bitszmll.
reg  [2:0] bitcnt;
wire       bitcnt_ld;
wire       bitcnt_tc = (bitcnt == 3'd0);

always @(posedge clk)
begin
   if (bitcnt_ld)
      case (data_bits)
         2'b00  : bitcnt <= 3'd4;  //5 bit
         2'b10  : bitcnt <= 3'd5;  //6 bit
         2'b01  : bitcnt <= 3'd6;  //7 bit
         default: bitcnt <= 3'd7;  //8 bit
      endcase
   else
      if (rx_shr_en)
         bitcnt <= bitcnt - 4'd1;
end


//******************************************************************************
//* Parits ellenrzs.                                                        *
//******************************************************************************
reg  par_err;
wire par_err_ld;

always @(posedge clk)
begin
   if (rst || (par_enable == 0))
      par_err <= 1'b0;
   else
      if (par_err_ld)
         case (par_type)
            2'b00  : par_err <=  ^{rxd_filtered, rx_data};  //Pros parits
            3'b01  : par_err <= ~^{rxd_filtered, rx_data};  //Pratlan parits
            3'b10  : par_err <= (rxd_filtered != 1'b0);     //Konstans 0
            default: par_err <= (rxd_filtered != 1'b1);     //Konstans 1
         endcase
end


//******************************************************************************
//* Vezrl llapotgp.                                                        *
//******************************************************************************
localparam RX_IDLE      = 3'd0;
localparam RX_STARTBIT  = 3'd1;
localparam RX_DATABITS  = 3'd2;
localparam RX_PARITYBIT = 3'd3;
localparam RX_STOPBIT1  = 3'd4;
localparam RX_STOPBIT2  = 3'd5;

reg [2:0] rx_state;

always @(posedge clk)
begin
   if (rst)
      rx_state <= RX_IDLE;
   else
      case (rx_state)
         //Vrakozs az engedlyezsre s a soros bemnet lefut lre.
         RX_IDLE     : if (enable & rxd_falling)
                          rx_state <= RX_STARTBIT;
                       else
                          rx_state <= RX_IDLE;
         
         //A START bit 0 rtknek ellenrzse.
         RX_STARTBIT : if (clkdiv_cnt_h_tc)
                          if (rxd_filtered)
                             rx_state <= RX_IDLE;
                          else
                             rx_state <= RX_DATABITS;
                       else
                          rx_state <= RX_STARTBIT;
                          
         //Az adatbitek vtele.
         RX_DATABITS : if (clkdiv_cnt_h_tc && bitcnt_tc)
                          if (par_enable)
                             rx_state <= RX_PARITYBIT;
                          else
                             rx_state <= RX_STOPBIT1;
                       else
                          rx_state <= RX_DATABITS;
         
         //A paritsbit vtele s ellenrzse.
         RX_PARITYBIT: if (clkdiv_cnt_h_tc)
                          rx_state <= RX_STOPBIT1;
                       else
                          rx_state <= RX_PARITYBIT;
         
         //Az els STOP bit vtele s ellenrzse.
         RX_STOPBIT1 : if (clkdiv_cnt_h_tc)
                          if (rxd_filtered && stop_bits)
                                rx_state <= RX_STOPBIT2;    
                          else
                             rx_state <= RX_IDLE;
                       else
                          rx_state <= RX_STOPBIT1;
         
         //A msodik STOP bit vtele s ellenrzse.
         RX_STOPBIT2 : if (clkdiv_cnt_h_tc)
                          rx_state <= RX_IDLE;
                       else
                          rx_state <= RX_STOPBIT2;
         
         //rvnytelen llapotok.
         default     : rx_state <= RX_IDLE;
      endcase
end

//A baud rate genertor engedlyez jele.
assign clkdiv_en  = ~(rx_state == RX_IDLE);
//A bitszmll betlt jele.
assign bitcnt_ld  =  (rx_state == RX_STARTBIT) & clkdiv_cnt_h_tc;
//A vteli shiftregiszter s a bitszmll engedlyez jele.
assign rx_shr_en  =  (rx_state == RX_DATABITS) & clkdiv_cnt_h_tc;
//A paritshiba jelz regiszter betlt jele.
assign par_err_ld =  (rx_state == RX_PARITYBIT) & clkdiv_cnt_h_tc;


//******************************************************************************
//* Vteli FIFO.                                                               *
//******************************************************************************
//A vteli FIFO rs engedlyez jele. A FIFO tele llapotot nem vizsgljuk,
//mert ez esetben a FIFO eldobja a bert adatot s az llapota nem mdosul.
reg rx_fifo_wr;

always @(*)
begin
   case (rx_state)
      RX_STOPBIT1: rx_fifo_wr <= rxd_filtered & clkdiv_cnt_h_tc & ~stop_bits;
      RX_STOPBIT2: rx_fifo_wr <= rxd_filtered & clkdiv_cnt_h_tc;
      default    : rx_fifo_wr <= 1'b0;
   endcase
end

//Vteli FIFO.
fifo #(
   //A FIFO mlysge.
   .DEPTH(16),
   //A szavak mrete bitekben.
   .WIDTH(9)
) rx_fifo (
   //rajel s reset.
   .clk(clk),                          //rajel.
   .rst(rst | fifo_clr),               //Reset jel.
   
   //Adat be- s kimenetek.
   .data_in({par_err, rx_data}),       //FIFO adat bemenet.
   .data_out(data_out),                //FIFO adat kimenet.
   
   //Vezrl bemenetek.
   .write(rx_fifo_wr),                 //FIFO rs engedlyez jel.
   .read(fifo_rd),                     //FIFO olvass engedlyez jel.
   
   //llapotjelz kimenetek.
   .exists(fifo_exists),               //A FIFO nem res.
   .full(fifo_full)                    //A FIFO tele van.
);

endmodule

`default_nettype wire
