`timescale 1ns / 1ps
`default_nettype none

//******************************************************************************
//* UART ad modul.                                                            *
//******************************************************************************
module uart_transmitter(
   //rajel s reset.
   input  wire        clk,             //Rendszerrajel
   input  wire        rst,             //Aktv magas szinkron reset
   
   //Ads engedlyets.
   input  wire        enable,
   
   //rajel oszts.
   input  wire [15:0] clk_div,
   
   //Vonali kdols.
   input  wire [1:0]  data_bits,       //Adatbitek szma
   input  wire        par_enable,      //Parits engedlyezs
   input  wire [1:0]  par_type,        //Parits tpusa
   input  wire        stop_bits,       //Stop bitek szma
   
   //Adsi FIFO.
   input  wire [7:0]  data_in,         //Adat bemenet
   input  wire        fifo_wr,         //rs engedlyez jel
   input  wire        fifo_clr,        //FIFO trl jel
   output wire        fifo_exists,     //A FIFO-ban van adat
   output wire        fifo_full,       //A FIFO tele van
   
   //Soros adat kimenet
   output wire        uart_txd
);

//******************************************************************************
//* Baud rate genertor.                                                       *
//******************************************************************************
reg  [19:0] clkdiv_cnt;
wire        clkdiv_en;
wire        clkdiv_cnt_tc = (clkdiv_cnt == 20'd0);

always @(posedge clk)
begin
   if ((clkdiv_en == 0) || clkdiv_cnt_tc)
      clkdiv_cnt <= {clk_div, 4'b1111};
   else
      clkdiv_cnt <= clkdiv_cnt - 20'd1;
end


//******************************************************************************
//* Adsi FIFO.                                                                *
//******************************************************************************
wire [7:0] tx_fifo_dout;
wire       tx_fifo_rd;

fifo #(
   //A FIFO mlysge.
   .DEPTH(16),
   //A szavak mrete bitekben.
   .WIDTH(8)
) tx_fifo (
   //rajel s reset.
   .clk(clk),                          //rajel.
   .rst(rst | fifo_clr),               //Reset jel.
   
   //Adat be- s kimenetek.
   .data_in(data_in),                  //FIFO adat bemenet.
   .data_out(tx_fifo_dout),            //FIFO adat kimenet.
   
   //Vezrl bemenetek.
   .write(fifo_wr),                    //FIFO rs engedlyez jel.
   .read(tx_fifo_rd),                  //FIFO olvass engedlyez jel.
   
   //llapotjelz kimenetek.
   .exists(fifo_exists),               //A FIFO nem res.
   .full(fifo_full)                    //A FIFO tele van.
);


//******************************************************************************
//* Parits genertor.                                                         *
//******************************************************************************
//Az adatmret alapjn a nem hasznlt bitek maszkolsa.
reg [7:0] parity_mask;

always @(*)
begin
   case (data_bits)
      2'b00  : parity_mask <= 8'b0001_1111;     //5 bit
      2'b10  : parity_mask <= 8'b0011_1111;     //6 bit
      2'b01  : parity_mask <= 8'b0111_1111;     //7 bit
      default: parity_mask <= 8'b1111_1111;     //8 bit
   endcase
end

//A maszkolt adatbitek XOR kapcsolata. Prosra egszti ki az 1 bitek szmt.
wire parity = ^(tx_fifo_dout & parity_mask);

//A paritsbit ellltsa a vonali kdolsi belltsok alapjn.
reg parity_out;

always @(*)
begin
   case ({par_enable, par_type})
      3'b100 : parity_out <=  parity;           //Pros parits
      3'b101 : parity_out <= ~parity;           //Pratlan parits
      3'b110 : parity_out <=  1'b0;             //Konstans 0
      3'b111 : parity_out <=  1'b1;             //Konstans 1
      default: parity_out <=  1'b1;             //Nincs engedlyezve
   endcase
end


//******************************************************************************
//* Adsi shiftregiszter.                                                      *
//******************************************************************************
//A shiftregiszterbe berand adat a vonali kdolsi belltsok alapjn.
reg [9:0] tx_shr_din;

always @(*)
begin
   case (data_bits)
      2'b00  : tx_shr_din <= {3'b111, parity_out, tx_fifo_dout[4:0], 1'b0};  //5 bit
      2'b10  : tx_shr_din <= {2'b11,  parity_out, tx_fifo_dout[5:0], 1'b0};  //6 bit
      2'b01  : tx_shr_din <= {1'b1,   parity_out, tx_fifo_dout[6:0], 1'b0};  //7 bit
      default: tx_shr_din <= {        parity_out, tx_fifo_dout[7:0], 1'b0};  //8 bit
   endcase
end

//Adsi shiftregiszter.
reg  [9:0] tx_shr;
wire       tx_shr_ld;
wire       tx_shr_en;

always @(posedge clk)
begin
   if (rst)
      tx_shr <= 10'b11_1111_1111;
   else
      if (tx_shr_ld)
         tx_shr <= tx_shr_din;
      else
         if (tx_shr_en)
            tx_shr <= {1'b1, tx_shr[9:1]};
end

assign uart_txd = tx_shr[0];

//Bitszmll.
reg  [3:0] bitcnt;
wire       bitcnt_tc = (bitcnt == 4'd0);

always @(posedge clk)
begin
   if (tx_shr_ld)
      case ({stop_bits, par_enable, data_bits})
         4'b0_0_00: bitcnt <= 4'd6;    //1 START + 5 adat + 0 parits + 1 STOP =  7 bit
         4'b0_0_01: bitcnt <= 4'd7;    //1 START + 6 adat + 0 parits + 1 STOP =  8 bit
         4'b0_0_10: bitcnt <= 4'd8;    //1 START + 7 adat + 0 parits + 1 STOP =  9 bit
         4'b0_0_11: bitcnt <= 4'd9;    //1 START + 8 adat + 0 parits + 1 STOP = 10 bit
         4'b0_1_00: bitcnt <= 4'd7;    //1 START + 5 adat + 1 parits + 1 STOP =  8 bit
         4'b0_1_01: bitcnt <= 4'd8;    //1 START + 6 adat + 1 parits + 1 STOP =  9 bit
         4'b0_1_10: bitcnt <= 4'd9;    //1 START + 7 adat + 1 parits + 1 STOP = 10 bit
         4'b0_1_11: bitcnt <= 4'd10;   //1 START + 8 adat + 1 parits + 1 STOP = 11 bit
         4'b1_0_00: bitcnt <= 4'd7;    //1 START + 5 adat + 0 parits + 2 STOP =  8 bit
         4'b1_0_01: bitcnt <= 4'd8;    //1 START + 6 adat + 0 parits + 2 STOP =  9 bit
         4'b1_0_10: bitcnt <= 4'd9;    //1 START + 7 adat + 0 parits + 2 STOP = 10 bit
         4'b1_0_11: bitcnt <= 4'd10;   //1 START + 8 adat + 0 parits + 2 STOP = 11 bit
         4'b1_1_00: bitcnt <= 4'd8;    //1 START + 5 adat + 1 parits + 2 STOP =  9 bit
         4'b1_1_01: bitcnt <= 4'd9;    //1 START + 6 adat + 1 parits + 2 STOP = 10 bit
         4'b1_1_10: bitcnt <= 4'd10;   //1 START + 7 adat + 1 parits + 2 STOP = 11 bit
         default  : bitcnt <= 4'd11;   //1 START + 8 adat + 1 parits + 2 STOP = 12 bit
      endcase
   else
      if (tx_shr_en)
         bitcnt <= bitcnt - 4'd1;
end


//******************************************************************************
//* Vezrl llapotgp.                                                        *
//******************************************************************************
localparam TX_IDLE = 2'd0;
localparam TX_INIT = 2'd1;
localparam TX_DATA = 2'd2;

reg [1:0] tx_state;

always @(posedge clk)
begin
   if (rst)
      tx_state <= TX_IDLE;
   else
      case (tx_state)
         //Vrakozs az engedlyezettsgre s a kldend adatra.
         TX_IDLE: if (enable && fifo_exists)
                     tx_state <= TX_INIT;
                  else
                     tx_state <= TX_IDLE;
         
         //A shiftregiszter s a bitszmll betltse.
         TX_INIT: tx_state <= TX_DATA;
         
         //Az ads vgrehajtsa.
         TX_DATA: if (clkdiv_cnt_tc && bitcnt_tc)
                     tx_state <= TX_IDLE;
                  else
                     tx_state <= TX_DATA;
         
         //rvnytelen llapotok.
         default: tx_state <= TX_IDLE;
      endcase
end

//A baud rate genertor engedlyez jele.
assign clkdiv_en  = (tx_state == TX_DATA);
//Az adsi FIFO olvass engedlyez jele.
assign tx_fifo_rd = (tx_state == TX_INIT);
//Az adsi shiftregiszter s a bitszmll betlt jele.
assign tx_shr_ld  = (tx_state == TX_INIT);
//Az adsi shiftregiszter s a bitszmll engedlyez jele.
assign tx_shr_en  = (tx_state == TX_DATA) & clkdiv_cnt_tc;

endmodule

`default_nettype wire
